-------------localisation-------------
--local L = AceLibrary("AceLocale-2.2"):new("GoodTarget")
--[[local L = LibStub("AceLocale-3.0"):GetLocale("GoodTarget")

L:RegisterTranslations("enUS", function()
	return {
		["GoodTarget"] = true,
	}
end)
]]
--------------Utility methods-------------

--cache string of the pattern "<unitid>target"
local UNITS_TARGET= setmetatable({}, {
	__index = function(t, i)
		local unitsTarget = i.."target"
		rawset(t, i, unitsTarget)
		return unitsTarget
	end,
}) 

-------------create Ace stuff-------------

--GoodTarget = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0")
GoodTarget = LibStub("AceAddon-3.0"):NewAddon("GoodTarget", "AceEvent-3.0", "AceTimer-3.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

---------------saved variables------------

GoodTarget.defaults = {
    CheckInterval = 0.05,
    PlaySound = true,
    isLocked = false,
    Scale = 1.0,
    Width = 100,
    Height = 20,
}

-------------------Options---------------
local options = { 
    type='group',
    handler = GoodTarget,
    args = {
        checkInterval = {
            type = 'range',
            name = 'Check interval',
            order = 2,
            desc = 'Set the interval in which GoodTarget checks the targets target (in seconds).',
            min = 0.01,
            max = 1.0,
            step = 0.01,
            bigStep = 0.02,            
            get = function() return GoodTarget.db.profile.CheckInterval end,
            set = "SetCheckInterval",
        },
        
        ---options---        
        playSound = {
            type = 'toggle',
            name = 'Play sounds',
            order = 20,
            desc = 'Enables or disables sounds on aggro loss or gain.',
            get = function() return GoodTarget.db.profile.PlaySound end,
            set = function(_, v) GoodTarget.db.profile.PlaySound = v end,
        },
        isLocked = {
            type = 'toggle',
            name = 'Lock frames',
            order = 21,
            desc = 'Locks or unlocks the frames.',
            get = function() return GoodTarget.db.profile.isLocked end,
            set = "SetLocked",
        },  
        
        ---Size---
        scale = {
            type = 'range',
            name = 'Scale',
            order = 60,
            desc = 'Set the scale of the frames.',
            min = 0.1,
            max = 4.0,
            step = 0.01,
            bigStep = 0.1,
            get = function() return GoodTarget.db.profile.Scale end,
            set = "SetScale",
        },
        width = {
            type = 'range',
            name = 'Width',
            order = 61,
            desc = 'Set the width of the frames.',
            min = 10,
            max = 500,
            step = 1,
            bigStep = 1,
            get = function() return GoodTarget.db.profile.Width end,
            set = "SetWidth",
        },
        height = {
            type = 'range',
            name = 'Height',
            order = 62,
            desc = 'Set the height of the frames.',
            min = 5,
            max = 100,
            step = 1,
            bigStep = 1,
            get = function() return GoodTarget.db.profile.Height end,
            set = "SetHeight",
        },
    },
}

AceConfig:RegisterOptionsTable("GoodTarget", options, "goodtarget")
----------options getter and setter-----------

---locked---
function GoodTarget:SetLocked(_, isLocked)
    self.db.profile.isLocked = isLocked
    self:LockFrame(isLocked)
    if self.lastPlayerTTType == self.NO_TARGET then
        self:SetNonePlayerFrame()
    end
end

---scale---
function GoodTarget:SetScale(_, scale)
    self.db.profile.Scale = scale
    self.playerFrame:SetScale(scale)
end

---width---
function GoodTarget:SetWidth(_, width)
    self.db.profile.Width = width
    self.playerFrame:SetWidth(width)
end

---height---
function GoodTarget:SetHeight(_, height)
    self.db.profile.Height = height
    self.playerFrame:SetHeight(height)
end

---check interval---
function GoodTarget:SetCheckInterval(_, interval)
    self.db.profile.CheckInterval = interval
    self:CancelTimer(self.gtTimer)
    self.gtTimer = self:ScheduleRepeatingTimer("GTCheckTargets", interval)
end


---------------initialisation--------------

function GoodTarget:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("GoodTargetDB", { profile = self.defaults }, true)
    
    -----constants-----
    self.NO_TARGET = 0
    self.GOOD_TARGET = 1
    self.BAD_TARGET = 2
    self.ALMOST_BAD_TARGET = 3
    
    -----init members----
    self.lastPlayerTTarget = "$none"
    self.lastPlayerTTType = self.NO_TARGET
    
    -----create frames-----
    self:CreatePlayerTargetFrame()
    
    -----init frames-----
    self:InitializeFrames()
    
    self.optFrame = AceConfigDialog:AddToBlizOptions("GoodTarget", "GoodTarget")
end

function GoodTarget:OnEnable()
    ---register events---
    self:RegisterEvent("PLAYER_TARGET_CHANGED")
    self:RegisterEvent("UNIT_THREAT_SITUATION_UPDATE", "GTCheckTargets")
    self.gtTimer = self:ScheduleRepeatingTimer("GTCheckTargets", self.db.profile.CheckInterval)
    
    
    ---show frames---
    --self:ShowFrames()
    
    ---update---
end

function GoodTarget:OnDisable()
    ---hide frames---
    self:HideFrames()
end

-----------------event handling-----------------
--[[Starts moving the frame on mouse down]]
function GoodTarget:OnMouseDown(f, button)
    if button == "LeftButton" then
		f:StartMoving()
	end
end

--[[Stops moving the frame on mouse up]]
function GoodTarget:OnMouseUp(f)
    f:StopMovingOrSizing()
end

--[[called whenever the current target of the player changes]]
function GoodTarget:PLAYER_TARGET_CHANGED()
    self:Update(true)
end

--[[This event is called in a specific interval and indicates
    to update the targets]]
function GoodTarget:GTCheckTargets()
    self:Update()
end


--------------actual functionality--------------

---general---
--[[Checks the targets and updates the frames]]
function GoodTarget:Update(playerTargetChanged)
    self:UpdatePlayerTTarget()
end

--[[Classifies the current tanking status of the passed unit by
    evaluating the units thread status on the player.]]
function GoodTarget:ClassifyUnitThread(unit)
    --show only tanking status for hostile units
    if not UnitExists(unit) or not UnitCanAttack("player", unit) then
        return self.NO_TARGET
    end
    
    local isTanking, status = UnitDetailedThreatSituation("player", unit)
    
    --fall back to classify tanking status by evaluating the units target
    if not status then
        return self:ClassifyUnitThreadByTarget(unit)
    end
    
    
    if status == 3 then --securely tanking
        return self.GOOD_TARGET
    elseif status == 2 then --almost loosing aggro
        return self.ALMOST_BAD_TARGET
    end
    --print("status: "..status)
    return self.BAD_TARGET
end




--[[Classifies the current tanking status of the passed unit by
    evaluating the units target.]]
function GoodTarget:ClassifyUnitThreadByTarget(unit)
    local unitsTarget = UNITS_TARGET[unit]       --units target
    
    --show only tanking status if unit has a target and it is our friend
    if not UnitExists(unitsTarget) or not UnitIsFriend("player", unitsTarget) then
        return self.NO_TARGET
    end
    
    --if the target of the unit is the player everything is fine
    if UnitIsUnit("player", unitsTarget) then
        return self.GOOD_TARGET
    end
    
    --otherwise player lost aggro
    --print("no status")
    return self.BAD_TARGET
end

---player target---
--[[Checks the current target of the player and updates the frame]]
function GoodTarget:UpdatePlayerTTarget()

    --self:Print("Update Player")
    local tTargetType = self:ClassifyPlayerTarget()
    
    if tTargetType == self.lastPlayerTTType and not self:PlayerTTargetChanged() then
        return
    end
    
    if tTargetType == self.GOOD_TARGET then
        --self:Print("good")
        self:SetGoodPlayerFrame()
        if self.db.profile.PlaySound then
            PlaySoundFile("Sound\\interface\\AuctionWindowOpen.ogg");
        end
    elseif tTargetType == self.ALMOST_BAD_TARGET then
        --self:Print("almost bad")
        self:SetAlmostBadFrame(self.playerFrame, "player")
    elseif  tTargetType == self.BAD_TARGET then
        --self:Print("bad")
        
        local unitsTT = "playertargettarget"
        local target = "NA"
        local color
        if UnitExists(unitsTT) then
            local _, class = UnitClass(unitsTT)
            color = RAID_CLASS_COLORS[class]
            target = UnitName(unitsTT)
        end
        self:SetBadPlayerFrame(target, color)
        if self.db.profile.PlaySound and tTargetType ~= self.lastPlayerTTType then
            PlaySoundFile("Sound\\Creature\\Cannon\\CannonDeath.ogg")
        end 
    else
        self:SetNonePlayerFrame()
    end
    
    self.lastPlayerTTType = tTargetType
end


--[[Checks if the targets target of the player has changed.
    Returns true if the players targets target has changed.]]
function GoodTarget:PlayerTTargetChanged()

    local tTarget = UnitGUID("playertargettarget")
    if not tTarget then
        tTarget = "$none"
    end
        
    if tTarget == self.lastPlayerTTarget then
        return false
    end
    
    self.lastPlayerTTarget = tTarget
    return true
end

--[[Classifies the targets target of the player as
    Good, bad or none.]]
function GoodTarget:ClassifyPlayerTarget()
    return self:ClassifyUnitThread("playertarget")
end


------------------GUI stuff-----------------

------frame creation------

---player frame---
function GoodTarget:CreatePlayerTargetFrame()
    local f=CreateFrame("FRAME","GoodTargetPlayerFrame",UIParent);
            
    f:SetBackdrop({
      bgFile="Interface\\Tooltips\\UI-Tooltip-Background", 
      tile=1, tileSize=10, edgeSize=0, 
      insets={left=0, right=0, top=0, bottom=0}
    });
    
    f:SetPoint("CENTER", UIParent)   
    f:SetMovable(true)
    
    f.master = self
    f:SetScript("OnMouseUp", function(f) f.master:OnMouseUp(f) end)
	f:SetScript("OnMouseDown", function(f, button) f.master:OnMouseDown(f, button) end)
    
    f.Text = f:CreateFontString("$parentText","ARTWORK","GameFontNormal");
    f.Text:SetPoint("CENTER",f)
    
    self.playerFrame = f
end

-----frame visibility-----

function GoodTarget:ShowFrames()
    self.playerFrame:Show()
end

function GoodTarget:HideFrames()
    self.playerFrame:Hide()
end


-----frame appeareance-----

---general---
function GoodTarget:InitializeFrames()
    local settings = self.db.profile
    
    self:LockFrame(settings.isLocked)
    
    local f = self.playerFrame
    f:SetWidth(settings.Width)
    f:SetHeight(settings.Height)
    f:SetScale(settings.Scale)
    self:SetNonePlayerFrame()

    self.playerFrame:Show()
end

function GoodTarget:LockFrame(locked)
    local f = self.playerFrame    
    f:EnableMouse(not locked)
end

---player frame---

function GoodTarget:SetGoodPlayerFrame()
    self:SetGoodFrame(self.playerFrame, "player")
end

function GoodTarget:SetBadPlayerFrame(target, color)
    self:SetBadFrame(self.playerFrame, "player", target, color)
end

function GoodTarget:SetNonePlayerFrame()
    if self.db.profile.isLocked then
        self:SetLockedPlayerFrame()
    else        
        self:SetUnlockedPlayerFrame()
    end
end

function GoodTarget:SetLockedPlayerFrame()
    self:SetNoneFrame(self.playerFrame)
end

function GoodTarget:SetUnlockedPlayerFrame()
    self.playerFrame:SetBackdropColor(0.5, 0.5, 0.5, 0.5)
    self.playerFrame.Text:SetTextColor(1, 1, 1, 1)
    self.playerFrame.Text:SetText("lock me to hide me")
end


function GoodTarget:SetBadFrame(frame, unit, target, color)    
    frame:SetBackdropColor(1, 0, 0, 1)
    --set name and color
    local text = frame.Text
    text:SetText(target or "NA")
    color = color or {r=1,g=1,b=1}
    text:SetTextColor(color.r, color.g, color.b, 1)       
end

function GoodTarget:SetAlmostBadFrame(frame, unit)
    frame:SetBackdropColor(0.8, 1, 0, 1)
    frame.Text:SetTextColor(1, 1, 1, 1)
    frame.Text:SetText("good...")  
end

function GoodTarget:SetGoodFrame(frame, unit)
    frame:SetBackdropColor(0, 1, 0, 1)
    frame.Text:SetTextColor(1, 1, 1, 1)
    frame.Text:SetText("good")     
end

function GoodTarget:SetNoneFrame(frame)
    frame:SetBackdropColor(1, 0, 0, 0)
    frame.Text:SetTextColor(1, 1, 1, 0)
end


